@@ -2,7 +2,6 @@ from __future__ import annotations
 
 import asyncio
 import os
-import platform
 import signal
 import sys
 import warnings
@@ -16,10 +15,13 @@ from typing import (
     AsyncGenerator,
     Awaitable,
     Callable,
+    cast,
     Coroutine,
     NoReturn,
     Optional,
+    overload,
     TypeVar,
+    Union,
 )
 from urllib.parse import quote
 from weakref import WeakSet
@@ -39,6 +41,7 @@ from werkzeug.wrappers import Response as WerkzeugResponse
 
 from .asgi import ASGIHTTPConnection, ASGILifespan, ASGIWebsocketConnection
 from .cli import AppGroup
+from .config import Config
 from .ctx import (
     _AppCtxGlobals,
     AppContext,
@@ -93,6 +96,7 @@ from .typing import (
     ASGIWebsocketProtocol,
     BeforeServingCallable,
     BeforeWebsocketCallable,
+    Event,
     FilePath,
     HeadersValue,
     ResponseReturnValue,
@@ -111,7 +115,6 @@ from .typing import (
 from .utils import (
     cancel_tasks,
     file_path_to_path,
-    is_coroutine_function,
     MustReloadError,
     observe_changes,
     restart,
@@ -173,6 +176,8 @@ class Quart(App):
             websocket protocol.
         config_class: The class to use for the configuration.
         env: The name of the environment the app is running on.
+        event_class: The class to use to signal an event in an async
+            manner.
         debug: Wrapper around configuration DEBUG value, in many places
             this will result in more output if True. If unset, debug
             mode will be activated if environ is set to 'development'.
@@ -187,6 +192,8 @@ class Quart(App):
         secret_key: Warpper around configuration SECRET_KEY value. The app
             secret for signing sessions.
         session_interface: The class to use as the session interface.
+        shutdown_event: This event is set when the app starts to
+            shutdown allowing waiting tasks to know when to stop.
         url_map_class: The class to map rules to endpoints.
         url_rule_class: The class to use for URL rules.
         websocket_class: The class to use for websockets.
@@ -196,6 +203,7 @@ class Quart(App):
     asgi_http_class: type[ASGIHTTPProtocol]
     asgi_lifespan_class: type[ASGILifespanProtocol]
     asgi_websocket_class: type[ASGIWebsocketProtocol]
+    shutdown_event: Event
     test_app_class: type[TestAppProtocol]
     test_client_class: type[TestClientProtocol]  # type: ignore[assignment]
 
@@ -204,6 +212,8 @@ class Quart(App):
     asgi_http_class = ASGIHTTPConnection
     asgi_lifespan_class = ASGILifespan
     asgi_websocket_class = ASGIWebsocketConnection
+    config_class = Config
+    event_class = asyncio.Event
     jinja_environment = Environment  # type: ignore[assignment]
     lock_class = asyncio.Lock
     request_class = Request
@@ -302,22 +312,22 @@ class Quart(App):
         )
 
         self.after_serving_funcs: list[Callable[[], Awaitable[None]]] = []
-        self.after_websocket_funcs: dict[
-            AppOrBlueprintKey, list[AfterWebsocketCallable]
-        ] = defaultdict(list)
+        self.after_websocket_funcs: dict[AppOrBlueprintKey, list[AfterWebsocketCallable]] = (
+            defaultdict(list)
+        )
         self.background_tasks: WeakSet[asyncio.Task] = WeakSet()
         self.before_serving_funcs: list[Callable[[], Awaitable[None]]] = []
-        self.before_websocket_funcs: dict[
-            AppOrBlueprintKey, list[BeforeWebsocketCallable]
-        ] = defaultdict(list)
-        self.teardown_websocket_funcs: dict[
-            AppOrBlueprintKey, list[TeardownCallable]
-        ] = defaultdict(list)
+        self.before_websocket_funcs: dict[AppOrBlueprintKey, list[BeforeWebsocketCallable]] = (
+            defaultdict(list)
+        )
+        self.teardown_websocket_funcs: dict[AppOrBlueprintKey, list[TeardownCallable]] = (
+            defaultdict(list)
+        )
         self.while_serving_gens: list[AsyncGenerator[None, None]] = []
 
         self.template_context_processors[None] = [_default_template_ctx_processor]
 
-        self.cli = AppGroup()  # type: ignore[assignment]
+        self.cli = AppGroup()
         self.cli.name = self.name
 
         if self.has_static_folder:
@@ -438,7 +448,7 @@ class Quart(App):
         extra_context: dict = {}
         for name in names:
             for processor in self.template_context_processors[name]:
-                extra_context.update(await self.ensure_async(processor)())
+                extra_context.update(await self.ensure_async(processor)())  # type: ignore
 
         original = context.copy()
         context.update(extra_context)
@@ -776,7 +786,8 @@ class Quart(App):
             warnings.warn(
                 f"Additional arguments, {','.join(kwargs.keys())}, are not supported.\n"
                 "They may be supported by Hypercorn, which is the ASGI server Quart "
-                "uses by default. This method is meant for development and debugging."
+                "uses by default. This method is meant for development and debugging.",
+                stacklevel=2,
             )
 
         if loop is None:
@@ -832,8 +843,6 @@ class Quart(App):
         print(f" * Running on {scheme}://{host}:{port} (CTRL + C to quit)")  # noqa: T201
 
         tasks = [loop.create_task(task)]
-        if platform.system() == "Windows":
-            tasks.append(loop.create_task(_windows_signal_support()))
 
         if use_reloader:
             tasks.append(loop.create_task(observe_changes(asyncio.sleep, shutdown_event)))
@@ -892,9 +901,9 @@ class Quart(App):
 
         return serve(self, config, shutdown_trigger=shutdown_trigger)
 
-    def test_client(self, use_cookies: bool = True) -> TestClientProtocol:
+    def test_client(self, use_cookies: bool = True, **kwargs: Any) -> TestClientProtocol:
         """Creates and returns a test client."""
-        return self.test_client_class(self, use_cookies=use_cookies)
+        return self.test_client_class(self, use_cookies=use_cookies, **kwargs)
 
     def test_cli_runner(self, **kwargs: Any) -> QuartCliRunner:
         """Creates and returns a CLI test runner."""
@@ -948,6 +957,26 @@ class Quart(App):
         self.after_websocket_funcs[None].append(func)
         return func
 
+    @setupmethod
+    def teardown_websocket(
+        self,
+        func: T_teardown,
+    ) -> T_teardown:
+        """Add a teardown websocket function.
+        This is designed to be used as a decorator, if used to
+        decorate a synchronous function, the function will be wrapped
+        in :func:`~quart.utils.run_sync` and run in a thread executor
+        (with the wrapped function returned). An example usage,
+        .. code-block:: python
+            @app.teardown_websocket
+            async def func():
+                ...
+        Arguments:
+            func: The teardown websocket function itself.
+        """
+        self.teardown_websocket_funcs[None].append(func)
+        return func
+
     async def handle_http_exception(
         self, error: HTTPException
     ) -> HTTPException | ResponseReturnValue:
@@ -972,7 +1001,7 @@ class Quart(App):
         if handler is None:
             return error
         else:
-            return await self.ensure_async(handler)(error)
+            return await self.ensure_async(handler)(error)  # type: ignore
 
     async def handle_user_exception(self, error: Exception) -> HTTPException | ResponseReturnValue:
         """Handle an exception that has been raised.
@@ -998,7 +1027,7 @@ class Quart(App):
         handler = self._find_error_handler(error, blueprints)
         if handler is None:
             raise error
-        return await self.ensure_async(handler)(error)
+        return await self.ensure_async(handler)(error)  # type: ignore
 
     async def handle_exception(self, error: Exception) -> ResponseTypes:
         """Handle an uncaught exception.
@@ -1029,7 +1058,7 @@ class Quart(App):
         handler = self._find_error_handler(server_error, request.blueprints)
 
         if handler is not None:
-            server_error = await self.ensure_async(handler)(server_error)
+            server_error = await self.ensure_async(handler)(server_error)  # type: ignore
 
         return await self.finalize_request(server_error, from_error_handler=True)
 
@@ -1061,7 +1090,7 @@ class Quart(App):
         handler = self._find_error_handler(server_error, websocket.blueprints)
 
         if handler is not None:
-            server_error = await self.ensure_async(handler)(server_error)
+            server_error = await self.ensure_async(handler)(server_error)  # type: ignore
 
         return await self.finalize_websocket(server_error, from_error_handler=True)
 
@@ -1084,7 +1113,15 @@ class Quart(App):
         else:
             self.logger.error("Exception", exc_info=exception_info)
 
-    def ensure_async(self, func: Callable[P, Any]) -> Callable[P, Awaitable[Any]]:
+    @overload
+    def ensure_async(self, func: Callable[P, Awaitable[T]]) -> Callable[P, Awaitable[T]]: ...
+
+    @overload
+    def ensure_async(self, func: Callable[P, T]) -> Callable[P, Awaitable[T]]: ...
+
+    def ensure_async(
+        self, func: Union[Callable[P, Awaitable[T]], Callable[P, T]]
+    ) -> Callable[P, Awaitable[T]]:
         """Ensure that the returned func is async and calls the func.
 
         .. versionadded:: 0.11
@@ -1093,10 +1130,10 @@ class Quart(App):
         run. Before Quart 0.11 this did not run the synchronous code
         in an executor.
         """
-        if is_coroutine_function(func):
+        if asyncio.iscoroutinefunction(func):
             return func
         else:
-            return self.sync_to_async(func)
+            return self.sync_to_async(cast(Callable[P, T], func))
 
     def sync_to_async(self, func: Callable[P, T]) -> Callable[P, Awaitable[T]]:
         """Return a async function that will run the synchronous function *func*.
@@ -1308,9 +1345,9 @@ class Quart(App):
         status: StatusCode | None = None
         if isinstance(result, tuple):
             if len(result) == 3:
-                value, status, headers = result  # type: ignore[misc]
+                value, status, headers = result
             elif len(result) == 2:
-                value, status_or_headers = result  # type: ignore[misc]
+                value, status_or_headers = result
 
                 if isinstance(status_or_headers, (Headers, dict, list)):
                     headers = status_or_headers
@@ -1323,7 +1360,7 @@ class Quart(App):
                     headers), or (body, status, headers)"""
                 )
         else:
-            value = result
+            value = result  # type: ignore[assignment]
 
         if value is None:
             raise TypeError("The response value returned by the view function cannot be None")
@@ -1339,7 +1376,7 @@ class Quart(App):
             ):
                 response = self.response_class(value)
             elif isinstance(value, (list, dict)):
-                response = self.json.response(value)
+                response = self.json.response(value)  # type: ignore[assignment]
             else:
                 raise TypeError(f"The response value type ({type(value).__name__}) is not valid")
         else:
@@ -1436,7 +1473,7 @@ class Quart(App):
             for function in self.before_request_funcs[name]:
                 result = await self.ensure_async(function)()
                 if result is not None:
-                    return result
+                    return result  # type: ignore
 
         return None
 
@@ -1462,7 +1499,7 @@ class Quart(App):
             for function in self.before_websocket_funcs[name]:
                 result = await self.ensure_async(function)()
                 if result is not None:
-                    return result
+                    return result  # type: ignore
 
         return None
 
@@ -1486,7 +1523,7 @@ class Quart(App):
             return await self.make_default_options_response()
 
         handler = self.view_functions[request_.url_rule.endpoint]
-        return await self.ensure_async(handler)(**request_.view_args)
+        return await self.ensure_async(handler)(**request_.view_args)  # type: ignore
 
     async def dispatch_websocket(
         self, websocket_context: WebsocketContext | None = None
@@ -1502,7 +1539,7 @@ class Quart(App):
             self.raise_routing_exception(websocket_)
 
         handler = self.view_functions[websocket_.url_rule.endpoint]
-        return await self.ensure_async(handler)(**websocket_.view_args)
+        return await self.ensure_async(handler)(**websocket_.view_args)  # type: ignore
 
     async def finalize_request(
         self,
@@ -1572,7 +1609,7 @@ class Quart(App):
         names = [*(request_context or request_ctx).request.blueprints, None]
 
         for function in (request_context or request_ctx)._after_request_functions:
-            response = await self.ensure_async(function)(response)
+            response = await self.ensure_async(function)(response)  # type: ignore
 
         for name in names:
             for function in reversed(self.after_request_funcs[name]):
@@ -1598,11 +1635,11 @@ class Quart(App):
         names = [*(websocket_context or websocket_ctx).websocket.blueprints, None]
 
         for function in (websocket_context or websocket_ctx)._after_websocket_functions:
-            response = await self.ensure_async(function)(response)
+            response = await self.ensure_async(function)(response)  # type: ignore
 
         for name in names:
             for function in reversed(self.after_websocket_funcs[name]):
-                response = await self.ensure_async(function)(response)
+                response = await self.ensure_async(function)(response)  # type: ignore
 
         session_ = (websocket_context or websocket_ctx).session
         if not self.session_interface.is_null_session(session_):
@@ -1647,6 +1684,7 @@ class Quart(App):
         await asgi_handler(receive, send)
 
     async def startup(self) -> None:
+        self.shutdown_event = self.event_class()
         try:
             async with self.app_context():
                 for func in self.before_serving_funcs:
@@ -1661,6 +1699,7 @@ class Quart(App):
             raise
 
     async def shutdown(self) -> None:
+        self.shutdown_event.set()
         try:
             await asyncio.wait_for(
                 asyncio.gather(*self.background_tasks),
@@ -1706,11 +1745,3 @@ def _cancel_all_tasks(loop: asyncio.AbstractEventLoop) -> None:
                     "task": task,
                 }
             )
-
-
-async def _windows_signal_support() -> None:
-    # See https://bugs.python.org/issue23057, to catch signals on
-    # Windows it is necessary for an IO event to happen periodically.
-    # Fixed by Python 3.8
-    while True:
-        await asyncio.sleep(1)
