@@ -1,13 +1,15 @@
 from __future__ import annotations
 
 import sys
+from contextvars import Token
 from functools import wraps
 from types import TracebackType
-from typing import Any, Callable, cast, Iterator, List, Optional, TYPE_CHECKING  # noqa: F401
+from typing import Any, Callable, cast, Iterator, List, Optional, Tuple, TYPE_CHECKING  # noqa: F401
 
+from flask.ctx import _AppCtxGlobals as _AppCtxGlobals  # noqa: F401
 from werkzeug.exceptions import HTTPException
 
-from .globals import _app_ctx_stack, _request_ctx_stack, _websocket_ctx_stack
+from .globals import _cv_app, _cv_request, _cv_websocket
 from .sessions import SessionMixin  # noqa
 from .signals import appcontext_popped, appcontext_pushed
 from .typing import AfterRequestCallable, AfterWebsocketCallable
@@ -31,27 +33,20 @@ class _BaseRequestWebsocketContext:
 
     def __init__(
         self,
-        app: "Quart",
+        app: Quart,
         request_websocket: BaseRequestWebsocket,
-        session: Optional[SessionMixin] = None,
+        session: SessionMixin | None = None,
     ) -> None:
         self.app = app
         self.request_websocket = request_websocket
         self.url_adapter = app.create_url_adapter(self.request_websocket)
         self.request_websocket.routing_exception = None
+        self.request_websocket.json_module = app.json
         self.session = session
         self.preserved = False
-        self._implicit_app_ctx_stack: List[Optional[AppContext]] = []
+        self._cv_tokens: list[tuple[Token, AppContext | None]] = []
 
-    @property
-    def g(self) -> AppContext:
-        return _app_ctx_stack.top.g
-
-    @g.setter
-    def g(self, value: AppContext) -> None:
-        _app_ctx_stack.top.g = value
-
-    def copy(self) -> "_BaseRequestWebsocketContext":
+    def copy(self) -> _BaseRequestWebsocketContext:
         return self.__class__(self.app, self.request_websocket, self.session)
 
     def match_request(self) -> None:
@@ -74,32 +69,33 @@ class _BaseRequestWebsocketContext:
     async def push(self) -> None:
         raise NotImplementedError()
 
-    async def pop(self, exc: Optional[BaseException]) -> None:
+    async def pop(self, exc: BaseException | None) -> None:
         raise NotImplementedError()
 
-    async def auto_pop(self, exc: Optional[BaseException]) -> None:
+    async def auto_pop(self, exc: BaseException | None) -> None:
         if self.request_websocket.scope.get("_quart._preserve_context", False) or (
-            exc is not None and self.app.preserve_context_on_exception
+            exc is not None and self.app.config["PRESERVE_CONTEXT_ON_EXCEPTION"]
         ):
             self.preserved = True
         else:
             await self.pop(exc)
 
-    async def __aenter__(self) -> "_BaseRequestWebsocketContext":
+    async def __aenter__(self) -> _BaseRequestWebsocketContext:
         await self.push()
         return self
 
     async def __aexit__(self, exc_type: type, exc_value: BaseException, tb: TracebackType) -> None:
         await self.auto_pop(exc_value)
 
-    async def _push_appctx(self) -> None:
-        app_ctx = _app_ctx_stack.top
-        if app_ctx is None or app_ctx.app != self.app:
+    async def _push_appctx(self, token: Token) -> None:
+        app_ctx = _cv_app.get(None)
+        if app_ctx is None or app_ctx.app is not self.app:
             app_ctx = self.app.app_context()
             await app_ctx.push()
-            self._implicit_app_ctx_stack.append(app_ctx)
         else:
-            self._implicit_app_ctx_stack.append(None)
+            app_ctx = None
+
+        self._cv_tokens.append((token, app_ctx))
 
     async def _push(self) -> None:
         if self.session is None:
@@ -129,37 +125,44 @@ class RequestContext(_BaseRequestWebsocketContext):
 
     def __init__(
         self,
-        app: "Quart",
+        app: Quart,
         request: Request,
-        session: Optional[SessionMixin] = None,
+        session: SessionMixin | None = None,
     ) -> None:
         super().__init__(app, request, session)
         self.flashes = None
-        self._after_request_functions: List[AfterRequestCallable] = []
+        self._after_request_functions: list[AfterRequestCallable] = []
 
     @property
     def request(self) -> Request:
         return cast(Request, self.request_websocket)
 
     async def push(self) -> None:
-        await super()._push_appctx()
-        _request_ctx_stack.push(self)
+        await super()._push_appctx(_cv_request.set(self))
         await super()._push()
 
-    async def pop(self, exc: Optional[BaseException] = _sentinel) -> None:  # type: ignore
-        app_ctx = self._implicit_app_ctx_stack.pop()
+    async def pop(self, exc: BaseException | None = _sentinel) -> None:  # type: ignore
         try:
-            if not self._implicit_app_ctx_stack:
-                self.preserved = False
+            if len(self._cv_tokens) == 1:
                 if exc is _sentinel:
                     exc = sys.exc_info()[1]
                 await self.app.do_teardown_request(exc, self)
+
+                request_close = getattr(self.request_websocket, "close", None)
+                if request_close is not None:
+                    await request_close()
         finally:
-            _request_ctx_stack.pop()
+            ctx = _cv_request.get()
+            token, app_ctx = self._cv_tokens.pop()
+            _cv_request.reset(token)
+
             if app_ctx is not None:
                 await app_ctx.pop(exc)
 
-    async def __aenter__(self) -> "RequestContext":
+            if ctx is not self:
+                raise AssertionError(f"Popped wrong request context. ({ctx!r} instead of {self!r})")
+
+    async def __aenter__(self) -> RequestContext:
         await self.push()
         return self
 
@@ -178,42 +181,44 @@ class WebsocketContext(_BaseRequestWebsocketContext):
 
     def __init__(
         self,
-        app: "Quart",
+        app: Quart,
         request: Websocket,
-        session: Optional[SessionMixin] = None,
+        session: SessionMixin | None = None,
     ) -> None:
         super().__init__(app, request, session)
-        self._after_websocket_functions: List[AfterWebsocketCallable] = []
+        self._after_websocket_functions: list[AfterWebsocketCallable] = []
 
     @property
     def websocket(self) -> Websocket:
         return cast(Websocket, self.request_websocket)
 
     async def push(self) -> None:
-        await super()._push_appctx()
-        _websocket_ctx_stack.push(self)
+        await super()._push_appctx(_cv_websocket.set(self))
         await super()._push()
 
-    async def pop(self, exc: Optional[BaseException] = _sentinel) -> None:  # type: ignore
-        app_ctx = self._implicit_app_ctx_stack.pop()
+    async def pop(self, exc: BaseException | None = _sentinel) -> None:  # type: ignore
         try:
-            if not self._implicit_app_ctx_stack:
-                self.preserved = False
+            if len(self._cv_tokens) == 1:
                 if exc is _sentinel:
                     exc = sys.exc_info()[1]
                 await self.app.do_teardown_websocket(exc, self)
         finally:
-            _websocket_ctx_stack.pop()
+            ctx = _cv_websocket.get()
+            token, app_ctx = self._cv_tokens.pop()
+            _cv_websocket.reset(token)
+
             if app_ctx is not None:
                 await app_ctx.pop(exc)
 
-    async def __aenter__(self) -> "WebsocketContext":
+            if ctx is not self:
+                raise AssertionError(f"Popped wrong request context. ({ctx!r} instead of {self!r})")
+
+    async def __aenter__(self) -> WebsocketContext:
         await self.push()
         return self
 
 
 class AppContext:
-
     """The context relating to the app bound to the current task.
 
     Do not use directly, prefer the
@@ -226,34 +231,37 @@ class AppContext:
         g: An instance of the ctx globals class.
     """
 
-    def __init__(self, app: "Quart") -> None:
+    def __init__(self, app: Quart) -> None:
         self.app = app
         self.url_adapter = app.create_url_adapter(None)
         self.g = app.app_ctx_globals_class()
-        self._app_reference_count = 0
+        self._cv_tokens: list[Token] = []
 
-    def copy(self) -> "AppContext":
+    def copy(self) -> AppContext:
         app_context = self.__class__(self.app)
         app_context.g = self.g
         return app_context
 
     async def push(self) -> None:
-        self._app_reference_count += 1
-        _app_ctx_stack.push(self)
-        await appcontext_pushed.send(self.app)
+        self._cv_tokens.append(_cv_app.set(self))
+        await appcontext_pushed.send_async(self.app, _sync_wrapper=self.app.ensure_async)
 
-    async def pop(self, exc: Optional[BaseException] = _sentinel) -> None:  # type: ignore
-        self._app_reference_count -= 1
+    async def pop(self, exc: BaseException | None = _sentinel) -> None:  # type: ignore
         try:
-            if self._app_reference_count <= 0:
+            if len(self._cv_tokens) == 1:
                 if exc is _sentinel:
                     exc = sys.exc_info()[1]
                 await self.app.do_teardown_appcontext(exc)
         finally:
-            _app_ctx_stack.pop()
-        await appcontext_popped.send(self.app)
+            ctx = _cv_app.get()
+            _cv_app.reset(self._cv_tokens.pop())
 
-    async def __aenter__(self) -> "AppContext":
+            if ctx is not self:
+                raise AssertionError(f"Popped wrong app context. ({ctx!r} instead of {self!r})")
+
+        await appcontext_popped.send_async(self.app, _sync_wrapper=self.app.ensure_async)
+
+    async def __aenter__(self) -> AppContext:
         await self.push()
         return self
 
@@ -277,7 +285,10 @@ def after_this_request(func: AfterRequestCallable) -> AfterRequestCallable:
 
             ...
     """
-    _request_ctx_stack.top._after_request_functions.append(func)
+    ctx = _cv_request.get(None)
+    if ctx is None:
+        raise RuntimeError("Not within a request context")
+    ctx._after_request_functions.append(func)
     return func
 
 
@@ -303,7 +314,10 @@ def after_this_websocket(func: AfterWebsocketCallable) -> AfterWebsocketCallable
             ...
 
     """
-    _websocket_ctx_stack.top._after_websocket_functions.append(func)
+    ctx = _cv_websocket.get(None)
+    if ctx is None:
+        raise RuntimeError("Not within a websocket context")
+    ctx._after_websocket_functions.append(func)
     return func
 
 
@@ -325,7 +339,7 @@ def copy_current_app_context(func: Callable) -> Callable:
     if not has_app_context():
         raise RuntimeError("Attempt to copy app context outside of a app context")
 
-    app_context = _app_ctx_stack.top.copy()
+    app_context = _cv_app.get().copy()
 
     @wraps(func)
     async def wrapper(*args: Any, **kwargs: Any) -> Any:
@@ -353,7 +367,7 @@ def copy_current_request_context(func: Callable) -> Callable:
     if not has_request_context():
         raise RuntimeError("Attempt to copy request context outside of a request context")
 
-    request_context = _request_ctx_stack.top.copy()
+    request_context = _cv_request.get().copy()
 
     @wraps(func)
     async def wrapper(*args: Any, **kwargs: Any) -> Any:
@@ -381,7 +395,7 @@ def copy_current_websocket_context(func: Callable) -> Callable:
     if not has_websocket_context():
         raise RuntimeError("Attempt to copy websocket context outside of a websocket context")
 
-    websocket_context = _websocket_ctx_stack.top.copy()
+    websocket_context = _cv_websocket.get().copy()
 
     @wraps(func)
     async def wrapper(*args: Any, **kwargs: Any) -> Any:
@@ -404,7 +418,7 @@ def has_app_context() -> bool:
 
     See also :func:`has_request_context`
     """
-    return _app_ctx_stack.top is not None
+    return _cv_app.get(None) is not None
 
 
 def has_request_context() -> bool:
@@ -420,7 +434,7 @@ def has_request_context() -> bool:
 
     See also :func:`has_app_context`.
     """
-    return _request_ctx_stack.top is not None
+    return _cv_request.get(None) is not None
 
 
 def has_websocket_context() -> bool:
@@ -436,50 +450,4 @@ def has_websocket_context() -> bool:
 
     See also :func:`has_app_context`.
     """
-    return _websocket_ctx_stack.top is not None
-
-
-class _AppCtxGlobals:
-    """The g class, a plain object with some mapping methods."""
-
-    def get(self, name: str, default: Optional[Any] = None) -> Any:
-        """Get a named attribute of this instance, or return the default."""
-        return self.__dict__.get(name, default)
-
-    def pop(self, name: str, default: Any = _sentinel) -> Any:
-        """Pop, get and remove the named attribute of this instance."""
-        if default is _sentinel:
-            return self.__dict__.pop(name)
-        else:
-            return self.__dict__.pop(name, default)
-
-    def setdefault(self, name: str, default: Any = None) -> Any:
-        """Set an attribute with a default value."""
-        return self.__dict__.setdefault(name, default)
-
-    def __contains__(self, item: Any) -> bool:
-        return item in self.__dict__
-
-    def __iter__(self) -> Iterator:
-        return iter(self.__dict__)
-
-    def __repr__(self) -> str:
-        top = _app_ctx_stack.top
-        if top is not None:
-            return f"<quart.g of {top.app.name}>"
-        return object.__repr__(self)
-
-    def __getattr__(self, name: str) -> Any:
-        try:
-            return self.__dict__[name]
-        except KeyError:
-            raise AttributeError(name) from None
-
-    def __setattr__(self, name: str, value: Any) -> None:
-        self.__dict__[name] = value
-
-    def __delattr__(self, name: str) -> None:
-        try:
-            del self.__dict__[name]
-        except KeyError:
-            raise AttributeError(name) from None
+    return _cv_websocket.get(None) is not None
